/** \file AnglesOptimization.h Declarations for the AnglesOptimization class.
    \author Liliya Kharevych
    \date March 2006
*/

#ifndef ANGLE_OPT_CM
#define ANGLE_OPT_CM
#ifdef _WIN32
#ifndef MSKDLL 
#define MSKDLL 
#endif
#endif
#include "mosek.h"
#include "Mesh.h"

/**
This class solves the angle optimization problem using quadratic programming. 

The variable of the quadratic program are \f$\lambda_i = \hat\alpha_i - \alpha_i\f$ and there are \f$3|F|\f$ variables, 
where \f$|F|\f$ is the number of triangles in the mesh. 
The quadratic objective is to minimize \f$\lambda_i^T Q \lambda_i\f$, where Q is just an identity matrix, 
and variables are subject to the following 4 types of constraints:

1. \f$
\forall i: \lambda_i \geq \epsilon-\alpha_i
\f$
Where we picked \f$\epsilon=10^{-3}\f$ to ensure that ">" relation is satisfied.

2. \f$
|F| 
\begin{array}{c} 3|F| \\
\left( \begin{array}{cccccccccc}
1 & 1 & 1 & 0 & 0 & 0 & 0 & 0 & 0 & \ldots \\
0 & 0 & 0 & 1 & 1 & 1 & 0 & 0 & 0 & \ldots \\
0 & 0 & 0 & 0 & 0 & 0 & 1 & 1 & 1 & \ldots \\
\vdots & \vdots & \vdots & \vdots & \vdots & \vdots & \vdots & \ddots 
\end{array} \right) \end{array}
\begin{array}{c}  \\ = \end{array}
\begin{array}{c} \\
\left( \begin{array}{c}
\pi - \sum_{i \in f_0} \alpha_i \\
\pi - \sum_{i \in f_1} \alpha_i \\
\pi - \sum_{i \in f_2} \alpha_i \\
\vdots
\end{array} \right) \end{array}
\f$

The sum of the current \f$\alpha\f$s in each triangle is subtracted from \f$\pi\f$ to 
ensure that \f$\lambda\f$s sum to the correct value, a similar subtraction is done in the next two matrices.

3.\f$
|V_c| 
\begin{array}{c} 3|F| \\
\left( \begin{array}{cccccccccc}
0 & 0 & 1 & 0 & 0 & 1 & 0 & 0 & 1 & \ldots \\
0 & 0 & 0 & 0 & 1 & 0 & 1 & 0 & 0 & \ldots \\
\vdots & \vdots & \vdots & \vdots & \vdots & \vdots & \vdots & \ddots 
\end{array} \right) \end{array}
\begin{array}{c}  \\ = \end{array}
\begin{array}{c}  \\
\left( \begin{array}{c}
2\pi - \sum_{i \prec v_0} \alpha_i  \\
\Theta - \sum_{i \prec v_1} \alpha_i  \\
\vdots
\end{array} \right) \end{array}
\f$

In the standard case, the sum of the angles around any internal vertex is fixed to \f$2\pi\f$ and the sum of the 
angles around any boundary vertex is left free.  However one can also fix boundary angles to control the shape of 
the parameter domain. It is also possible to fix the sum of the angles around some internal vertices to 
\f$\Theta \neq 2\pi\f$, which makes this vertices cone singularities, while the rest of the parameterization remains 
continuous piecewise flat surface. In this case \f$|V_c|\f$ is the number of vertices, for which the cone angle is 
fixed (it is also possible to fix the range of the sum).

4. \f$
|E_{int}|
\begin{array}{c} 3|F| \\
\left( \begin{array}{cccccccccc}
1 & 0 & 0 & 0 & 1 & 0 & 0 & 0 & 0 & \ldots \\
0 & 0 & 0 & 1 & 0 & 0 & 0 & 1 & 0 & \ldots \\
\vdots & \vdots & \vdots & \vdots & \vdots & \vdots & \vdots & \ddots 
\end{array} \right) \end{array}
\begin{array}{c}  \\ \leq \end{array}
\begin{array}{c} \\
\left( \begin{array}{c}
\pi - \alpha_l - \alpha_r - \epsilon \\
\pi - \alpha_l - \alpha_r - \epsilon\\
\vdots
\end{array} \right) \end{array}
\f$

Where \f$|E_{int}|\f$ is the number of internal edges, because the previous constraints ensure that \f$\theta\f$ for the 
boundary edge are within the valid bounds. The {0,1} values in all the 
above matrices correspond to incidence relations is described in the following figure.
\image html numberedAlphas.png
*/


class AnglesOptimization
{
private:

   Mesh * mesh;

   MSKtask_t task;

   /** Number of constraints.             */
   int NUMCON;
   /** Number of variables.               */
   int NUMVAR;
   /** Number of nonzeros in A (constraint matrix).          */
   int NUMANZ;
   /** Number of nonzeros in Q (quadratic objective matrix). */
   int NUMQNZ;      

   /** Type of the constraint                    

   Size: NUMCON
   */
   int *bkc;

   /** Lower bound on the constraint                    

   Size: NUMCON
   */
   double * blc;

   /** Upper bound on the constraint                    

   Size: NUMCON
   */
   double * buc;

   /** For representation of sparse constraint matrix: beginning of the row                     

   Size: NUMVAR
   */
   int *ptrb;

   /** For representation of sparse constraint matrix: ending of the row

   Size: NUMVAR
   */
   int *ptre;

   /** For representation of sparse constraint matrix: subscripts for nonzero values in the row                    

   Size: NUMANZ
   */
   int *sub;

   /** For representation of sparse constraint matrix: values in the matrix                     

   Size: NUMANZ
   */
   double * val;


   /** Type of the constraint on the variable    

   Size: NUMVAR
   */
   int *bkx;

   /** Lower bound on the variable               

   Size: NUMVAR
   */
   double * blx;

   /** Upper bound on the variable             

   Size: NUMVAR
   */
   double * bux;


   /** Linear minimization part                

   Size: NUMVAR
   */
   double * c;

   /** Quadratic objective matrix j subscript 

   Size: NUMQNZ
   */
   int *qsubi;

   /** Quadratic objective matrix i subscript 

   Size: NUMQNZ
   */
   int *qsubj;

   /** Quadratic objective matrix non-zero values 

   The lower triangular part.
   Size: NUMQNZ
   */
   double * qval; 


   /** Variable that we optimize for - alpha angles          

   Size: NUMVAR
   */
   double * xx;

public:

   /** Value of epsilon: how far can the values of alphas be from 0 or PI. */
   static const double epsRange;

   /*! \brief Creates angle optimization task.
   *  
   * Computes number of variables, constraints, 
   * number of non-zeros in constraint and objective matrix.
   */
   AnglesOptimization(Mesh * _mesh, const MSKenv_t & env);

   void doSolve();

   /** Sets linear constraints and quadratic objective */
   void setDataStructres();

   /** Calls MOSEK functions to optimize the task and save the solution */
   void optimize();

   /** Alocates arrays and matrices required by Mosek */
   void allocateMosekArrays();

   /** Deletes arrays and matrices required by Mosek */
   void clearMosekArrays();

   /** Clears mosek task */
   ~AnglesOptimization(void) { clearMosekArrays(); }

};
#endif

