/** \file CirclePattern.cpp Definitions for the CirclePattern structure.
    \author Liliya Kharevych
    \date March 2006
*/

#include "CirclePattern.h"
#include "util.h"
#include "Mesh.h"

using namespace std;

static double M2_PI = 2*M_PI;

/**
Computes 
\f$\textrm{ ImLi}_2(e^{\rho_k - \rho_l+i\theta}) + \textrm{ ImLi}_2(e^{\rho_l - \rho_k+i\theta})\f$,
where \f$x=\rho_k - \rho_l\f$.
\param theta            valid theta angle for edge e
\param x                \f$=\rho_k - \rho_l=\log(r_k) - \log(r_l)\f$
\param tanHalfThetaStar \f$=\tan(\frac{\pi - \theta}{2})\f$
\param clThetaStar      \f$=\textrm{Cl}_2(2(\pi - \theta))\f$.
*/
double CirclePattern::ImLi2Sum (const double& theta, const double& x, 
				const double& tanHalfThetaStar, const double& clThetaStar)
{
   double tStar = M_PI - theta;
   double p = 2*atan(tanh(0.5*x)*tanHalfThetaStar);

   return  p*x + Cl2(p + tStar) + Cl2(-p + tStar) - clThetaStar;
}

/** Computes circle pattern energy for the mesh.


The energy is computed as: 
\f$\sum_{E_{int}} (ImLi2Sum(\theta, x_i - x_j) - (\pi - \theta)(x_i - x_j)) - \sum_{E_{bdry}} 2 x_i (\pi-\theta)
+ \sum_{F}2\pi x_f\f$

*/
double CirclePattern::evaluateEnergy (Mesh * mesh, double *xx) {
   int i, j;
   double energy = 0;
   for (i = 0; i < mesh->numFaces(); i++) {
      energy += xx[i];
   }

   energy = M2_PI*energy;   // Total contribution from faces

   Mesh::HalfEdgeIterator hIter = mesh->halfEdgeIterator();
   for (hIter.reset(); !hIter.end(); hIter++) {
      Edge * e = hIter.half_edge();

      if (!e->isBoundary()) {
	 i = e->face->ID;
	 j = e->pair->face->ID;
	 if (i > j) {
	    double theta = e->theta();
	    energy += ImLi2Sum(theta, xx[i] - xx[j], e->tanHalfTStar(), e->Cl2TStar());
	    energy -= (M_PI - theta)*(xx[i] + xx[j]);
	 }
      }
      else {
	 if (e->face) {
	    i = e->face->ID;
	    energy -= 2*(M_PI - e->theta())*(xx[i]);
	 }
      }
   }
   return energy;
}

/**
Computes gradient vector for circle pattern energy.

For each triangle t: \f$\textrm{ Grad}(t) = 2\pi - \sum_{e \in t}2\varphi_e^t\f$,

where for internal edge e: \f$\varphi_e^t=\frac{\pi}{2}-\textrm{ atan2}(1 - e^x \cos\theta_e ,e^x \sin\theta_e)\f$, with \f$x = xx_i - xx_j\f$; 
and  for boundary edge e: \f$\varphi_e^t=\pi-\theta_e\f$.
*/
void CirclePattern::computeGradient (Mesh * mesh, double * xx, double * grad) {

   Mesh::FaceIterator fIter = mesh->faceIterator();

   /* Compute and store the gradient of the f. */
   int i, j;
   for (i = 0; i < mesh->numFaces(); i++)
      grad[i] = M2_PI;

   // sum over edges
   Mesh::HalfEdgeIterator hIter = mesh->halfEdgeIterator();
   for (hIter.reset(); !hIter.end(); hIter++) {
      Edge * e = hIter.half_edge();
      if (!e->isBoundary()) {
	 i = e->face->ID;
	 j = e->pair->face->ID;
	 if (i > j) {
	    grad[i] -= fTheta2(e->cosTheta(), e->sinTheta(), xx[j] - xx[i]);
	    grad[j] -= fTheta2(e->cosTheta(), e->sinTheta(), xx[i] - xx[j]);
	 }
      }
      else if (e->face) {
	 i = e->face->ID;
	 grad[i] -= 2*(M_PI - e->theta());
      }
   }

}

/** Sets the sparsity pattern of the non-zeros in the Hessian. */
void CirclePattern::buildSparsityForHessian (Mesh * mesh, int * hessI, int * hessJ) {
   int i, k, j;

   /** Non-zeros on the diagonal */
   for (k = 0; k < mesh->numFaces(); k++) {
      hessI[k] = k; 
      hessJ[k] = k;
   }

   k = mesh->numFaces();
   /** And non zero element [i][j] for each internal edge (i,j) */
   Mesh::HalfEdgeIterator hIter = mesh->halfEdgeIterator();
   for (hIter.reset(); !hIter.end(); hIter++) {
      if (!hIter.half_edge()->isBoundary()) {
	 i = hIter.half_edge()->face->ID;
	 j  = hIter.half_edge()->pair->face->ID;
	 if (i > j) {
	    hessI[k] = i; 
	    hessJ[k] = j;
	    k++;
	 }       
      }
   }
}

/** Computes the sparse Hessian of circle pattern energy. 

The value of Hessian[i][j] = 
Hessian [j,i] = -\f$\frac{\partial \varphi_e^t(x)}{\partial x} = -\frac {\sin(\theta)}{\cosh(x) - \cos(\theta)}\f$;
The value of Hessian[i][i] = \f$\sum_{e \in t_i} \frac{\partial \varphi_e^t(x)}{\partial x}\f$,
with \f$x = xx_i - xx_j\f$.
*/
void CirclePattern::computeHessian (Mesh * mesh, double * xx, double * hessval) {
   for (int k = 0; k < mesh->numFaces(); k++) 
      hessval[k] = 0;

   int count = mesh->numFaces();
   int i,j;
   Mesh::HalfEdgeIterator hIter = mesh->halfEdgeIterator();
   for (hIter.reset(); !hIter.end(); hIter++) {
      Edge * e = hIter.half_edge();
      if (!e->isBoundary()) {
	 i = e->face->ID;
	 j = e->pair->face->ID;
	 if (i > j) {
	    double ximj = xx[i] - xx[j];
	    double fThPrime2 = fThetaPrime2(e->cosTheta(), e->sinTheta(), ximj);
	    hessval[count] = -fThPrime2;
	    hessval[i] += fThPrime2;
	    hessval[j] += fThPrime2;
	    count++;
	 }
      }
   }
}

