/** \file CirclePattern.h Declarations for the CirclePattern structure.
    \author Liliya Kharevych
    \date March 2006
*/
#ifndef CIRCLE_PATTERN_CM
#define CIRCLE_PATTERN_CM
#define NOMINMAX
#include <iostream>
#define _USE_MATH_DEFINES
#include <math.h>

class Mesh;

/** Defines functions that compute quantities relevant to circle pattern energy 

EnergyMinimization class calls these functions while solving the non-linear problem. 

The most important functions are evaluateEnergy, computeGradient, computeHessian, and buildSparsityForHessian.
These functions can be modified to use different non-linear solver and different datastructures for gradient and Hessian.

The energy is computed as: 

\f$S_{ecl}(x)=\sum_{E_{int}}^{(i,j)} [ \textrm{ ImLi2Sum }(\theta, x_i - x_j) - (\pi - \theta)(x_i - x_j)]\f$
\f$- \sum_{E_{bdry}}^{(i,j)} 2 x_i (\pi-\theta) + \sum_{F}2\pi x_f\f$

The gradient is:

\f$\frac{\partial S_{ecl}(x)}{\partial x_i} = 2\pi - \sum_{E_{int}\in f_i}^{(i,j)}2 \arctan \frac{\sin\theta}{e^{x_i-x_j} - \cos\theta}\f$
\f$- \sum_{E_{bdry} \in f_i}^{(i,j)} \pi-\theta_e\f$

The Hessian is:

\f$\frac{\partial^2 S_{ecl}(x)}{\partial x_i \partial x_j} = -\frac {\sin(\theta)}{\cosh(x_i-x_j) - \cos(\theta)}\f$

\f$\frac{\partial^2 S_{ecl}(x)}{\partial x_i \partial x_i} = \sum_{E_{int}\in f_i}^{(i,j)} \frac {\sin(\theta)}{\cosh(x_i-x_j) - \cos(\theta)}\f$

All the sums are taken over edges (not the half-edges). 
The variables x represent logarithm of the radii of the triangles in the mesh. 
The variables \f$\theta\f$ are defined for each edge and represent an angle of intersection of two 
circumcircles of triangles meeting on this edge.

*/
struct CirclePattern {

   static double evaluateEnergy (Mesh * mesh, double * xx);

   static void computeGradient (Mesh * mesh, double * xx, double * grad);

   static void buildSparsityForHessian (Mesh * mesh, int * hessI, int * hessJ);

   static void computeHessian (Mesh * mesh, double * xx, double * hessval);


   static double ImLi2Sum (const double& theta, const double& x, 
      const double& tanHalfThetaStar, const double& clThetaStar);

   /** Implementation of Clausen integral taken from Boris Springborn's code                                   */
   static double Cl2 (const double & _x);

   /** Computes 
   \f$2\varphi_e=2\textrm{ atan2}(\sin\theta_e, e^{r_2-r_1} - \cos\theta_e)=\pi-2\textrm{ atan2}(1-e^{r_2-r_1}\cos\theta, e^{r_2-r_1}\sin\theta)\f$ */
   static double fTheta2(const double& cosTheta, const double& sinTheta, const double& x);

   /** Computes \f$2\frac{\partial \varphi_e^t(x)}{\partial x} = 2\frac {\sin(\theta)}{\cosh(x) - \cos(\theta)}\f$ */
   static double fThetaPrime2(const double & cosTheta, const double& sinTheta, const double & x) {
      double res = sinTheta/(cosh(x) - cosTheta);
      return res;
   }

   /** Reimplementation of java version of IEEEremainder, when x is divided by div                            */
   static double IEEEremainder(const double & x, const double & div);
};

inline double CirclePattern::fTheta2(const double& cosTheta, const double& sinTheta, const double& x){
   double ex = exp(x);
   double s1 = 1.0 - ex*cosTheta;
   double s2 = ex*sinTheta;

   double res = M_PI - 2*atan2(s1, s2);

   return res;
}

inline double CirclePattern::IEEEremainder(const double & x, const double & div) {

   double  res;
   if (x >= 0) {
      if (x < div) res = x;
      else {
	 double part = floor(x / div);
	 res = x - part*div;
      }
   }
   else {
      double part = floor(x / div);
      res = x - part*div;
   }
   if (res < 0 || res > div) {
      std::cout << "x " << x << " div " << div << std::endl;	
      std::cerr << "Wrong IEEEreminder!!!" << std::endl;
   }
   return res;
}

inline double CirclePattern::Cl2 (const double & _x) {
   double TWO_PI = 2 * M_PI;
   double MINUS_LOG_2 = log(0.5);
   double BORDERLINE = 2.0944;

   if (_x == 0.0) {
      return 0.0;
   }

   double x = IEEEremainder(_x, TWO_PI);

   if (x == 0.0) {
      return 0.0;
   }

   if (fabs(x) <= BORDERLINE) {
      double xx = x * x;
      return ((((((((((((
	 2.3257441143020875e-22 * xx
	 + 1.0887357368300848e-20) * xx 
	 + 5.178258806090624e-19) * xx
	 + 2.5105444608999545e-17) * xx
	 + 1.2462059912950672e-15) * xx
	 + 6.372636443183181e-14) * xx
	 + 3.387301370953521e-12) * xx
	 + 1.8978869988971e-10) * xx
	 + 1.1482216343327455e-8) * xx
	 + 7.873519778281683e-7) * xx
	 + 0.00006944444444444444) * xx
	 + 0.013888888888888888) * xx 
	 - log(fabs(x)) + 1.0) * x;
   }
   x += ((x > 0.0) ? - M_PI : M_PI);
   double xx = x * x;
   return ((((((((((((  
      3.901950904063069e-15 * xx
      + 4.566487567193635e-14) * xx
      + 5.429792727596476e-13) * xx
      + 6.5812165661369675e-12) * xx
      + 8.167010963952222e-11) * xx
      + 1.0440290284867003e-9) * xx
      + 1.3870999114054669e-8) * xx
      + 1.941538399871733e-7) * xx
      + 2.927965167548501e-6) * xx
      + 0.0000496031746031746) * xx
      + 0.0010416666666666667) * xx
      + 0.041666666666666664) * xx 
      + MINUS_LOG_2) * x;
}

#endif

