/** \file Edge.h Declarations and definitions for the Edge class.
    \author Liliya Kharevych
    \date March 2006
*/


#ifndef EDGE_CM
#define EDGE_CM
#include "Vector3.h"
#include "CirclePattern.h"

class Vertex;
class Face;

/** Half edge for the mesh datastructure.

In addition to the usual members which edge for the half-edge datastructure has 
(such as pointers to the next and pair edges, pointer to the vertex at the start 
of the edge and pointer to the touching face); 
this edge class has some extra data for circle pattern algorithm. 
The main data is a length of an edge and a theta angle associated with an edge. 
We also store an alpha angle associated with each edge (aplhaOposite). 
This angle is formed by another two edges of the triangle which touches the edge. 
For boundary edges alphaOposite are equal to 0.
Various functions of theta angle are precomputed and stored: 
cos(theta), sin(theta), tan(0.5*(M_PI - theta)),
Cl2(2.0*(M_PI - theta)).


*/
class Edge
{
private:
	/** theta angles */
   double th;
   double cTheta;
   double sTheta;
   double cl2TS;
   double tanHTS;

public:

   // pointers for mesh structure
   Edge * next;
   Edge * pair;
   Vertex * vertex;
   Face * face;

   /** Alpha angle oposite to the edge (different for pair edges). */
   double alphaOposite;

   /** Check for various iterations*/
   bool check;

   /**Length of the edge (same for pair edges). */
   double length;

   /**Used for angle optimization */
   int ID;

   /**If the mesh is cut into multiple patches, 
   this stores the patch ID that edge belongs to. */
   int patchID;


   Edge(void) : next(NULL), pair(NULL), vertex(NULL), face(NULL),
      length(0)
   {
      patchID = 0;
      th = 0;
      alphaOposite = 0;
   }

   // Assignment (copy only geometrical information)
   void assignData(const Edge& rhs){
      if (this != &rhs) {
	 alphaOposite = rhs.alphaOposite;
	 length = rhs.length; 
	 th = rhs.th; 
	 cTheta = rhs.cTheta; 
	 sTheta = rhs.sTheta; 
	 cl2TS = rhs.cl2TS; 
	 tanHTS = rhs.tanHTS;
      }
   }

   bool isBoundary() const {return !(face && pair->face);}
   Vertex * oppositeVertex() {return next->next->vertex;}

   const double & Cl2TStar() const {return cl2TS;}
   const double & tanHalfTStar() const {return tanHTS;}
   const double & sinTheta() const {return sTheta;}
   const double & cosTheta() const {return cTheta;}

   const double & theta() const {return th;}
   void setTheta(const double & _th) {
      th = _th;
      cTheta = cos(th);
      sTheta = sin(th);
      tanHTS = tan(0.5*(M_PI - th));
      cl2TS = CirclePattern::Cl2(2.0*(M_PI - th));   
   }

};
#endif

