/** \file EnergyMinimization.h Declarations for the EnergyMinimization class.
    \author Liliya Kharevych
    \date March 2006
*/

#ifndef ENERGY_MIN_CM
#define ENERGY_MIN_CM
#ifdef _WIN32
#ifndef MSKDLL 
#define MSKDLL 
#endif
#endif
#include "mosek.h"
#include "Mesh.h"

/** This class uses MOSEK library to solve non-linear convex minimization problem.

In order to find flat (or piecewise flat) triangulation of the input surface, valid radii
of the circumcircles of the triangles need to be computed. This is done by minimizing convex functional.
This class creates MOSEK minimization task that uses functions defined in CirclePattern class 
to do the minimization.
*/
class EnergyMinimization
{
private:
   Mesh * mesh;

   MSKtask_t task;

   /**
   Define constraints and free bounds for the variables.
   The only constraint for the optimization \f$\sum_{f \in F}\rho_f = 0\f$ (removes scaling direction).
   For most of the non-linear solvers this constraint is not necessary, as they would not move in the
   direction that does not change the energy anyway.
   */
   void setDataStructres();

   /** Calls MOSEK functions to perform minimization. */
   void minimize();

   /** Allocate arrays that MOSEK uses for minimization. */
   void allocateMosekArrays();

   /** Delete arrays that MOSEK uses for minimization. */
   void clearMosekArrays();

   /** Number of constraints.             */
   int NUMCON;  

   /** Number of variables.               */
   int NUMVAR;

   /** Number of nonzeros i A.            */
   int NUMANZ;    

   /** Type of the constraint                    

   Size: NUMCON   */
   int * bkc;

   /** Lower bound on the constraint                    

   Size: NUMCON   */
   double  * blc;

   /** Upper bound on the constraint                    

   Size: NUMCON   */
   double  * buc;

   /** For represenations of sparse constraint matrix: beginning of the row                     

   Size: NUMVAR   */
   int *ptrb;

   /** For represenations of sparse constraint matrix: ending of the row

   Size: NUMVAR   */
   int *ptre;

   /** For represenations of sparse constraint matrix: subscipts for nonzero values in the row                    

   Size: NUMANZ   */
   int *sub;

   /** For represenations of sparse constraint matrix: values in the matrix                     

   Size: NUMANZ   */
   double * val;

   /** Type of the constraint on the variable    

   Size: NUMVAR   */
   int * bkx;

   /** Lower bound on the variable               

   Size: NUMVAR   */
   double  * blx;

   /** Upper bound on the variable             

   Size: NUMVAR   */

   double  * bux;

   /** Linear minimization part                

   Size: NUMVAR   */
   double * c;

   /** Variable that we optimize for - radii of circum centers of triangles 

   Size: NUMVAR   */
   double  * xx;


public:

   /*! \brief Creates circle pattern energy minimization task.
   *  
   * Computes number of variables, constraints, 
   * number of non-zeros in constraint matrix, 
   * sets functions to compute energy, gradient and hessian.
   */
   EnergyMinimization(Mesh * _mesh, const MSKenv_t & env);

   /** This function calls rutines to perform circle pattern energy minimization*/
   void doSolve();

   /** Delete MOSEK arrays allocated for minimization */
   ~EnergyMinimization(void) { clearMosekArrays(); }

};

#endif

