/** \file util.h
File contains various utility functions.
*/
#ifndef UTIL_CM
#define UTIL_CM
#include <iostream>
#include <string>

static void check_error (bool expr, const std::string & message);
static void check_warn (bool expr, const std::string & message);
static void usage();

/** If expression \a expr is true, \a message is output and program terminates.
\param expr: Exression to check
\param message: Error message to output.
*/
static void check_error (bool expr, const std::string & message) {
   if (expr) {
      std::cerr << "Error: " << message << std::endl;
      exit(1);
   }
}

/** If expression \a expr is true, warning \a message is output.
\param expr: Exression to check
\param message: Warning message to output.
*/
static void check_warn (bool expr, const std::string & message) {
   if (expr)
      std::cerr << "Warning: " << message << std::endl;
}

//static void MSKAPI printMosekStr(void *handle, char str[]) {
//   std::ostream * mosekOut = (std::ostream *) handle;
//   check_error (!mosekOut, "No output stream specified for MOSEK.");
//   *mosekOut << str; 
//} 

/** 
Usage printout function.
*/
static void usage() {
   std::cerr << "USAGE:\n    CirclePatterns.exe <INPUT_OPTIONS> "
      << " <OUTPUT_OPTIONS> [OTHER OPTIONS] | <INPUT_FILE_WITH_OPTIONS>\n" << std::endl;
   std::cerr << "OPTIONS:\n"
      << "  MAIN OPTIONS:\n"
      << "    -io <input obj file>\n"
      << "    -oo <output obj file>\n"
      << "    <text file with options>\n"
      << "  ADDITIONAL INPUT OPTIONS:\n"
      << "    -ic <input conn file>\n"
      << "    -iv <singularities file>\n"
      << "    -ie <edges to cut file>\n"
      << "  ADDITIONAL OUTPUT OPTIONS:\n"
      << "    -oc <output conn file>\n"
      << "    -ob <output cut obj file>\n"
      << "    -ovt <otput vt file>\n"
      << "  OTHER OPTIONS:\n"
      << "    -no (only cut and layout mesh, no parameterization is computed)\n"
      << "    -help (outputs usage message)\n"
      << std::endl;

   exit(1);
}

/** 
Structure to store command line options
input by user
*/
struct ProgramOptions {
   /** Input (OBJ or CON) file */
   std::string file_in;

   /** Input file with edges to cut */
   std::string edge_file;

   /** Input file with list of cone singularities */
   std::string vert_file;

   /** Output file (CON format). */
   std::string conn_file_out;

   /** Output file (OBJ format). */
   std::string obj_file_out;

   /** Output file (VT format). */
   std::string vt_file_out;

   /** True if OBJ input file specified. */
   bool read_obj;

   /** True if CON input file specified. */
   bool read_conn;

   /** True if cone singularities are given. */
   bool read_vert;

   /** True if cut edges are specified. */
   bool read_edges;

   /** True if OBJ file needs to be written. */
   bool write_obj;

   /** True if VT file needs to be written. */
   bool write_vt;

   /** True if CON file needs to be written. */
   bool write_conn;

   /** True when no optimization needs to be performed.
   Can be useful when piecewise flat surface need to be cut.*/
   bool do_optim;

   ProgramOptions(): file_in(""), edge_file(""), vert_file(""),
      conn_file_out(""), obj_file_out(""), vt_file_out(""),
      read_obj(false), read_conn(false), 
      read_vert(false), read_edges(false),
      write_obj(false), write_conn(false),
      write_vt(false), do_optim(true)
   {}
};

#endif

